@@ -31,6 +31,7 @@
 #include <libxml/parser.h>
 #include <libxml/tree.h>
 
+#include "empathy-tp-chat.h"
 #include "empathy-chatroom-manager.h"
 #include "empathy-utils.h"
 
@@ -39,10 +40,16 @@
 
 #define CHATROOMS_XML_FILENAME "chatrooms.xml"
 #define CHATROOMS_DTD_FILENAME "empathy-chatroom-manager.dtd"
+#define SAVE_TIMER 4
+
+static EmpathyChatroomManager *chatroom_manager_singleton = NULL;
 
 #define GET_PRIV(obj) EMPATHY_GET_PRIV (obj, EmpathyChatroomManager)
 typedef struct {
 	GList      *chatrooms;
+  gchar *file;
+  /* source id of the autosave timer */
+  gint save_timer_id;
 } EmpathyChatroomManagerPriv;
 
 static void     chatroom_manager_finalize          (GObject                    *object);
@@ -52,6 +59,7 @@ static gboolean chatroom_manager_file_parse        (EmpathyChatroomManager
 static void     chatroom_manager_parse_chatroom    (EmpathyChatroomManager      *manager,
 						    xmlNodePtr                  node);
 static gboolean chatroom_manager_file_save         (EmpathyChatroomManager      *manager);
+static void reset_save_timeout (EmpathyChatroomManager *self);
 
 enum {
 	CHATROOM_ADDED,
@@ -61,15 +69,118 @@ enum {
 
 static guint signals[LAST_SIGNAL];
 
+/* properties */
+enum
+{
+  PROP_FILE = 1,
+  LAST_PROPERTY
+};
+
 G_DEFINE_TYPE (EmpathyChatroomManager, empathy_chatroom_manager, G_TYPE_OBJECT);
 
+static void
+empathy_chatroom_manager_get_property (GObject *object,
+                                       guint property_id,
+                                       GValue *value,
+                                       GParamSpec *pspec)
+{
+  EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (object);
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
+
+  switch (property_id)
+    {
+      case PROP_FILE:
+        g_value_set_string (value, priv->file);
+        break;
+      default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+        break;
+    }
+}
+
+static void
+empathy_chatroom_manager_set_property (GObject *object,
+                                       guint property_id,
+                                       const GValue *value,
+                                       GParamSpec *pspec)
+{
+  EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (object);
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
+
+  switch (property_id)
+    {
+      case PROP_FILE:
+        g_free (priv->file);
+        priv->file = g_value_dup_string (value);
+        break;
+      default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+        break;
+    }
+}
+
+static GObject *
+empathy_chatroom_manager_constructor (GType type,
+                                      guint n_props,
+                                      GObjectConstructParam *props)
+{
+  GObject *obj;
+  EmpathyChatroomManager *self;
+  EmpathyChatroomManagerPriv *priv;
+
+  if (chatroom_manager_singleton != NULL)
+    return g_object_ref (chatroom_manager_singleton);
+
+  /* Parent constructor chain */
+  obj = G_OBJECT_CLASS (empathy_chatroom_manager_parent_class)->
+        constructor (type, n_props, props);
+
+  self = EMPATHY_CHATROOM_MANAGER (obj);
+  priv = GET_PRIV (self);
+
+  chatroom_manager_singleton = self;
+  g_object_add_weak_pointer (obj, (gpointer *) &chatroom_manager_singleton);
+
+  if (priv->file == NULL)
+    {
+      /* Set the default file path */
+      gchar *dir;
+
+      dir = g_build_filename (g_get_home_dir (), ".gnome2", PACKAGE_NAME, NULL);
+      if (!g_file_test (dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
+        g_mkdir_with_parents (dir, S_IRUSR | S_IWUSR | S_IXUSR);
+
+      priv->file = g_build_filename (dir, CHATROOMS_XML_FILENAME, NULL);
+      g_free (dir);
+    }
+
+  chatroom_manager_get_all (self);
+  return obj;
+}
+
 static void
 empathy_chatroom_manager_class_init (EmpathyChatroomManagerClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
+  GParamSpec *param_spec;
 
+  object_class->constructor = empathy_chatroom_manager_constructor;
+  object_class->get_property = empathy_chatroom_manager_get_property;
+  object_class->set_property = empathy_chatroom_manager_set_property;
 	object_class->finalize = chatroom_manager_finalize;
 
+  param_spec = g_param_spec_string (
+      "file",
+      "path of the favorite file",
+      "The path of the XML file containing user's favorites",
+      NULL,
+      G_PARAM_CONSTRUCT_ONLY |
+      G_PARAM_READWRITE |
+      G_PARAM_STATIC_NAME |
+      G_PARAM_STATIC_NICK |
+      G_PARAM_STATIC_BLURB);
+  g_object_class_install_property (object_class, PROP_FILE, param_spec);
+
 	signals[CHATROOM_ADDED] =
 		g_signal_new ("chatroom-added",
 			      G_TYPE_FROM_CLASS (klass),
@@ -102,37 +213,89 @@ empathy_chatroom_manager_init (EmpathyChatroomManager *manager)
 	manager->priv = priv;
 }
 
+static void
+chatroom_changed_cb (EmpathyChatroom *chatroom,
+                     GParamSpec *spec,
+                     EmpathyChatroomManager *self)
+{
+  reset_save_timeout (self);
+}
+
 static void
 chatroom_manager_finalize (GObject *object)
 {
+  EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (object);
 	EmpathyChatroomManagerPriv *priv;
+  GList *l;
 
 	priv = GET_PRIV (object);
 
-	g_list_foreach (priv->chatrooms, (GFunc) g_object_unref, NULL);
+  if (priv->save_timer_id > 0)
+    {
+      /* have to save before destroy the object */
+      g_source_remove (priv->save_timer_id);
+      priv->save_timer_id = 0;
+      chatroom_manager_file_save (self);
+    }
+
+  for (l = priv->chatrooms; l != NULL; l = g_list_next (l))
+    {
+      EmpathyChatroom *chatroom = l->data;
+
+      g_signal_handlers_disconnect_by_func (chatroom, chatroom_changed_cb,
+          self);
+
+      g_object_unref (chatroom);
+    }
+
 	g_list_free (priv->chatrooms);
+  g_free (priv->file);
 
 	(G_OBJECT_CLASS (empathy_chatroom_manager_parent_class)->finalize) (object);
 }
 
 EmpathyChatroomManager *
-empathy_chatroom_manager_new (void)
+empathy_chatroom_manager_dup_singleton (const gchar *file)
 {
-	static EmpathyChatroomManager *manager = NULL;
+	return EMPATHY_CHATROOM_MANAGER (g_object_new (EMPATHY_TYPE_CHATROOM_MANAGER,
+		"file", file, NULL));
+}
 
-	if (!manager) {
-		EmpathyChatroomManagerPriv *priv;
+static gboolean
+save_timeout (EmpathyChatroomManager *self)
+{
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
 
-		manager = g_object_new (EMPATHY_TYPE_CHATROOM_MANAGER, NULL);
-		priv = GET_PRIV (manager);
-		chatroom_manager_get_all (manager);
-	
-		g_object_add_weak_pointer (G_OBJECT (manager), (gpointer) &manager);
-	} else {
-		g_object_ref (manager);
-	}
+  priv->save_timer_id = 0;
+  chatroom_manager_file_save (self);
 
-	return manager;
+  return FALSE;
+}
+
+static void
+reset_save_timeout (EmpathyChatroomManager *self)
+{
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
+
+  if (priv->save_timer_id > 0)
+    {
+      g_source_remove (priv->save_timer_id);
+    }
+
+  priv->save_timer_id = g_timeout_add_seconds (SAVE_TIMER,
+      (GSourceFunc) save_timeout, self);
+}
+
+static void
+add_chatroom (EmpathyChatroomManager *self,
+              EmpathyChatroom *chatroom)
+{
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
+
+  priv->chatrooms = g_list_prepend (priv->chatrooms, g_object_ref (chatroom));
+
+  g_signal_connect (chatroom, "notify",
+      G_CALLBACK (chatroom_changed_cb), self);
 }
 
 gboolean
@@ -150,8 +313,16 @@ empathy_chatroom_manager_add (EmpathyChatroomManager *manager,
 	if (!empathy_chatroom_manager_find (manager,
 					   empathy_chatroom_get_account (chatroom),
 					   empathy_chatroom_get_room (chatroom))) {
-		priv->chatrooms = g_list_prepend (priv->chatrooms, g_object_ref (chatroom));
-		chatroom_manager_file_save (manager);
+      gboolean favorite;
+
+      g_object_get (chatroom, "favorite", &favorite, NULL);
+
+    add_chatroom (manager, chatroom);
+
+    if (favorite)
+      {
+        reset_save_timeout (manager);
+      }
 
 		g_signal_emit (manager, signals[CHATROOM_ADDED], 0, chatroom);
 
@@ -178,12 +349,23 @@ empathy_chatroom_manager_remove (EmpathyChatroomManager *manager,
 
 		this_chatroom = l->data;
 
-		if (empathy_chatroom_equal (chatroom, this_chatroom)) {
+		if (this_chatroom == chatroom ||
+        empathy_chatroom_equal (chatroom, this_chatroom)) {
+        gboolean favorite;
 			priv->chatrooms = g_list_delete_link (priv->chatrooms, l);
 
-			chatroom_manager_file_save (manager);
+      g_object_get (chatroom, "favorite", &favorite, NULL);
+
+      if (favorite)
+        {
+          reset_save_timeout (manager);
+        }
 
 			g_signal_emit (manager, signals[CHATROOM_REMOVED], 0, this_chatroom);
+
+      g_signal_handlers_disconnect_by_func (chatroom, chatroom_changed_cb,
+          manager);
+
 			g_object_unref (this_chatroom);
 			break;
 		}
@@ -283,14 +465,6 @@ empathy_chatroom_manager_get_count (EmpathyChatroomManager *manager,
 	return count;
 }
 
-void
-empathy_chatroom_manager_store (EmpathyChatroomManager *manager)
-{
-	g_return_if_fail (EMPATHY_IS_CHATROOM_MANAGER (manager));
-
-	chatroom_manager_file_save (manager);
-}
-
 /*
  * API to save/load and parse the chatrooms file.
  */
@@ -299,27 +473,13 @@ static gboolean
 chatroom_manager_get_all (EmpathyChatroomManager *manager)
 {
 	EmpathyChatroomManagerPriv *priv;
-	gchar                     *dir;
-	gchar                     *file_with_path = NULL;
 
 	priv = GET_PRIV (manager);
 
-	dir = g_build_filename (g_get_home_dir (), ".gnome2", PACKAGE_NAME, NULL);
-	if (!g_file_test (dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
-		g_mkdir_with_parents (dir, S_IRUSR | S_IWUSR | S_IXUSR);
-	}
-
-	file_with_path = g_build_filename (dir, CHATROOMS_XML_FILENAME, NULL);
-	g_free (dir);
-
 	/* read file in */
-	if (g_file_test (file_with_path, G_FILE_TEST_EXISTS) &&
-	    !chatroom_manager_file_parse (manager, file_with_path)) {
-		g_free (file_with_path);
-		return FALSE;
-	}
-
-	g_free (file_with_path);
+	if (g_file_test (priv->file, G_FILE_TEST_EXISTS) &&
+	    !chatroom_manager_file_parse (manager, priv->file))
+    return FALSE;
 
 	return TRUE;
 }
@@ -433,7 +593,8 @@ chatroom_manager_parse_chatroom (EmpathyChatroomManager *manager,
 	}
 
 	chatroom = empathy_chatroom_new_full (account, room, name, auto_connect);
-	priv->chatrooms = g_list_prepend (priv->chatrooms, chatroom);
+  g_object_set (chatroom, "favorite", TRUE, NULL);
+  add_chatroom (manager, chatroom);
 	g_signal_emit (manager, signals[CHATROOM_ADDED], 0, chatroom);
 
 	g_object_unref (account);
@@ -449,19 +610,9 @@ chatroom_manager_file_save (EmpathyChatroomManager *manager)
 	xmlDocPtr                  doc;
 	xmlNodePtr                 root;
 	GList                     *l;
-	gchar                     *dir;
-	gchar                     *file;
 
 	priv = GET_PRIV (manager);
 
-	dir = g_build_filename (g_get_home_dir (), ".gnome2", PACKAGE_NAME, NULL);
-	if (!g_file_test (dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
-		g_mkdir_with_parents (dir, S_IRUSR | S_IWUSR | S_IXUSR);
-	}
-
-	file = g_build_filename (dir, CHATROOMS_XML_FILENAME, NULL);
-	g_free (dir);
-
 	doc = xmlNewDoc ("1.0");
 	root = xmlNewNode (NULL, "chatrooms");
 	xmlDocSetRootElement (doc, root);
@@ -470,8 +621,14 @@ chatroom_manager_file_save (EmpathyChatroomManager *manager)
 		EmpathyChatroom *chatroom;
 		xmlNodePtr      node;
 		const gchar    *account_id;
+    gboolean favorite;
 
 		chatroom = l->data;
+
+    g_object_get (chatroom, "favorite", &favorite, NULL);
+    if (!favorite)
+      continue;
+
 		account_id = mc_account_get_unique_name (empathy_chatroom_get_account (chatroom));
 
 		node = xmlNewChild (root, NULL, "chatroom", NULL);
@@ -484,14 +641,101 @@ chatroom_manager_file_save (EmpathyChatroomManager *manager)
 	/* Make sure the XML is indented properly */
 	xmlIndentTreeOutput = 1;
 
-	DEBUG ("Saving file:'%s'", file);
-	xmlSaveFormatFileEnc (file, doc, "utf-8", 1);
+	DEBUG ("Saving file:'%s'", priv->file);
+	xmlSaveFormatFileEnc (priv->file, doc, "utf-8", 1);
 	xmlFreeDoc (doc);
 
 	xmlCleanupParser ();
 	xmlMemoryDump ();
 
-	g_free (file);
-
 	return TRUE;
 }
+
+static void
+chatroom_manager_chat_destroyed_cb (EmpathyTpChat *chat,
+  gpointer user_data)
+{
+  EmpathyChatroomManager *manager = EMPATHY_CHATROOM_MANAGER (user_data);
+  McAccount *account = empathy_tp_chat_get_account (chat);
+  EmpathyChatroom *chatroom;
+  const gchar *roomname;
+  gboolean favorite;
+
+  roomname = empathy_tp_chat_get_id (chat);
+  chatroom = empathy_chatroom_manager_find (manager, account, roomname);
+
+  if (chatroom == NULL)
+    return;
+
+  g_object_set (chatroom, "tp-chat", NULL, NULL);
+  g_object_get (chatroom, "favorite", &favorite, NULL);
+
+  if (!favorite)
+    {
+      /* Remove the chatroom from the list, unless it's in the list of
+       * favourites..
+       * FIXME this policy should probably not be in libempathy */
+      empathy_chatroom_manager_remove (manager, chatroom);
+    }
+}
+
+static void
+chatroom_manager_observe_channel_cb (EmpathyDispatcher *dispatcher,
+  EmpathyDispatchOperation *operation, gpointer user_data)
+{
+  EmpathyChatroomManager *manager = EMPATHY_CHATROOM_MANAGER (user_data);
+  EmpathyChatroom *chatroom;
+  TpChannel *channel;
+  EmpathyTpChat *chat;
+  const gchar *roomname;
+  GQuark channel_type;
+  TpHandleType handle_type;
+  McAccount *account;
+
+  channel_type = empathy_dispatch_operation_get_channel_type_id (operation);
+
+  /* Observe Text channels to rooms only */
+  if (channel_type != TP_IFACE_QUARK_CHANNEL_TYPE_TEXT)
+    return;
+
+  channel = empathy_dispatch_operation_get_channel (operation);
+  tp_channel_get_handle (channel, &handle_type);
+
+  if (handle_type != TP_HANDLE_TYPE_ROOM)
+    return;
+
+  chat = EMPATHY_TP_CHAT (
+    empathy_dispatch_operation_get_channel_wrapper (operation));
+  account = empathy_tp_chat_get_account (chat);
+
+  roomname = empathy_tp_chat_get_id (chat);
+
+  chatroom = empathy_chatroom_manager_find (manager, account, roomname);
+
+  if (chatroom == NULL)
+    {
+      chatroom = empathy_chatroom_new_full (account, roomname, roomname,
+        FALSE);
+      g_object_set (G_OBJECT (chatroom), "tp-chat", chat, NULL);
+      empathy_chatroom_manager_add (manager, chatroom);
+      g_object_unref (chatroom);
+    }
+  else
+    {
+      g_object_set (G_OBJECT (chatroom), "tp-chat", chat, NULL);
+    }
+
+  /* A TpChat is always destroyed as it only gets unreffed after the channel
+   * has been invalidated in the dispatcher..  */
+  g_signal_connect (chat, "destroy",
+    G_CALLBACK (chatroom_manager_chat_destroyed_cb),
+    manager);
+}
+
+void
+empathy_chatroom_manager_observe (EmpathyChatroomManager *manager,
+  EmpathyDispatcher *dispatcher)
+{
+  g_signal_connect (dispatcher, "observe",
+    G_CALLBACK (chatroom_manager_observe_channel_cb), manager);
+}
