@@ -27,8 +27,11 @@
 #include <glade/glade.h>
 #include <gdk/gdkkeysyms.h>
 
+#include <libnotify/notification.h>
+
 #include <libempathy/empathy-utils.h>
 #include <libempathy/empathy-idle.h>
+#include <libempathy/empathy-account-manager.h>
 
 #include <libempathy-gtk/empathy-presence-chooser.h>
 #include <libempathy-gtk/empathy-conf.h>
@@ -40,6 +43,7 @@
 #include "empathy-status-icon.h"
 #include "empathy-preferences.h"
 #include "empathy-event-manager.h"
+#include "empathy-misc.h"
 
 #define DEBUG_FLAG EMPATHY_DEBUG_DISPATCHER
 #include <libempathy/empathy-debug.h>
@@ -51,12 +55,12 @@
 typedef struct {
 	GtkStatusIcon       *icon;
 	EmpathyIdle         *idle;
-	MissionControl      *mc;
+	EmpathyAccountManager *account_manager;
 	gboolean             showing_event_icon;
 	guint                blink_timeout;
-	gpointer             token;
 	EmpathyEventManager *event_manager;
 	EmpathyEvent        *event;
+	NotifyNotification  *notification;
 
 	GtkWindow           *window;
 	GtkWidget           *popup_menu;
@@ -67,21 +71,120 @@ typedef struct {
 
 G_DEFINE_TYPE (EmpathyStatusIcon, empathy_status_icon, G_TYPE_OBJECT);
 
+static gboolean
+activate_event (EmpathyEvent *event)
+{
+	empathy_event_activate (event);
+
+	return FALSE;
+}
+
+static void
+status_icon_notification_closed_cb (NotifyNotification *notification,
+				    EmpathyStatusIcon  *icon)
+{
+	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
+	EmpathyNotificationClosedReason reason = 0;
+
+#ifdef notify_notification_get_closed_reason
+	reason = notify_notification_get_closed_reason (notification);
+#endif
+	if (priv->notification) {
+		g_object_unref (priv->notification);
+		priv->notification = NULL;
+	}
+
+	if (!priv->event) {
+		return;
+	}
+
+	/* the notification has been closed by the user, see the
+	 * DesktopNotification spec.
+	 */
+	if (reason == EMPATHY_NOTIFICATION_CLOSED_DISMISSED) {
+		/* use an idle here, as this callback is called from a
+		 * DBus signal handler inside libnotify, and we might call
+		 * a *_run_* method when activating the event.
+		 */
+		g_idle_add ((GSourceFunc) activate_event, priv->event);
+	} else {
+		/* inhibit other updates for this event */
+		empathy_event_inhibit_updates (priv->event);
+	}
+}
+
+static void
+notification_close_helper (EmpathyStatusIconPriv *priv)
+{
+	if (priv->notification) {
+		notify_notification_close (priv->notification, NULL);
+		g_object_unref (priv->notification);
+		priv->notification = NULL;
+	}
+}
+
+static void
+status_icon_update_notification (EmpathyStatusIcon *icon)
+{
+	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
+	GdkPixbuf *pixbuf = NULL;
+
+	if (!empathy_notification_is_enabled ()) {
+		/* always close the notification if this happens */
+		notification_close_helper (priv);
+		return;
+	}
+
+	if (priv->event) {
+		pixbuf = empathy_misc_get_pixbuf_for_notification (priv->event->contact,
+								   priv->event->icon_name);
+
+		if (priv->notification) {
+			notify_notification_update (priv->notification,
+						    priv->event->header, priv->event->message,
+						    NULL);
+		} else {
+			priv->notification = notify_notification_new_with_status_icon
+				(priv->event->header, priv->event->message, NULL, priv->icon);
+			notify_notification_set_timeout (priv->notification,
+							 NOTIFY_EXPIRES_DEFAULT);
+
+			g_signal_connect (priv->notification, "closed",
+					  G_CALLBACK (status_icon_notification_closed_cb), icon);
+
+ 		}
+		notify_notification_set_icon_from_pixbuf (priv->notification,
+							  pixbuf);
+		notify_notification_show (priv->notification, NULL);
+
+		g_object_unref (pixbuf);
+	} else {
+		notification_close_helper (priv);
+	}
+}
+
 static void
 status_icon_update_tooltip (EmpathyStatusIcon *icon)
 {
 	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
-	const gchar           *tooltip = NULL;
+	gchar                 *tooltip = NULL;
 
 	if (priv->event) {
-		tooltip = priv->event->message;
+		tooltip = g_strdup_printf ("%s\n%s",
+					   priv->event->header,
+					   priv->event->message);
 	}
 
 	if (!tooltip) {
-		tooltip = empathy_idle_get_status (priv->idle);
+		tooltip = g_strdup (empathy_idle_get_status (priv->idle));
 	}
 
-	gtk_status_icon_set_tooltip (priv->icon, tooltip);	
+	/* FIXME: when we will depend on GTK+ 2.16.0, we should use
+	 * gtk_status_icon_set_tooltip_markup () and make the header italic.
+	 */
+	gtk_status_icon_set_tooltip (priv->icon, tooltip);
+
+	g_free (tooltip);
 }
 
 static void
@@ -112,7 +215,6 @@ status_icon_blink_timeout_cb (EmpathyStatusIcon *icon)
 
 	return TRUE;
 }
-
 static void
 status_icon_event_added_cb (EmpathyEventManager *manager,
 			    EmpathyEvent        *event,
@@ -131,6 +233,7 @@ status_icon_event_added_cb (EmpathyEventManager *manager,
 
 	status_icon_update_icon (icon);
 	status_icon_update_tooltip (icon);
+	status_icon_update_notification (icon);
 
 	if (!priv->blink_timeout) {
 		priv->blink_timeout = g_timeout_add (BLINK_TIMEOUT,
@@ -155,12 +258,35 @@ status_icon_event_removed_cb (EmpathyEventManager *manager,
 	status_icon_update_tooltip (icon);
 	status_icon_update_icon (icon);
 
+	/* update notification anyway, as it's safe and we might have been
+	 * changed presence in the meanwhile
+	 */	
+	status_icon_update_notification (icon);
+
 	if (!priv->event && priv->blink_timeout) {
 		g_source_remove (priv->blink_timeout);
 		priv->blink_timeout = 0;
 	}
 }
 
+static void
+status_icon_event_updated_cb (EmpathyEventManager *manager,
+			      EmpathyEvent        *event,
+			      EmpathyStatusIcon   *icon)
+{
+	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
+
+	if (event != priv->event) {
+		return;
+	}
+
+	if (empathy_notification_is_enabled ()) {
+		status_icon_update_notification (icon);
+	}
+
+	status_icon_update_tooltip (icon);
+}
+
 static void
 status_icon_set_visibility (EmpathyStatusIcon *icon,
 			    gboolean           visible,
@@ -217,8 +343,20 @@ status_icon_toggle_visibility (EmpathyStatusIcon *icon)
 static void
 status_icon_idle_notify_cb (EmpathyStatusIcon *icon)
 {
+	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
+
 	status_icon_update_icon (icon);
 	status_icon_update_tooltip (icon);
+
+	if (!empathy_notification_is_enabled ()) {
+		/* dismiss the outstanding notification if present */
+
+		if (priv->notification) {
+			notify_notification_close (priv->notification, NULL);
+			g_object_unref (priv->notification);
+			priv->notification = NULL;
+		}
+	}
 }
 
 static gboolean
@@ -341,6 +479,23 @@ status_icon_create_menu (EmpathyStatusIcon *icon)
 	g_object_unref (glade);
 }
 
+static void
+status_icon_connection_changed_cb (EmpathyAccountManager *manager,
+				   McAccount *account,
+				   TpConnectionStatusReason reason,
+				   TpConnectionStatus current,
+				   TpConnectionStatus previous,
+				   EmpathyStatusIcon *icon)
+{
+	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
+	int connected_accounts;
+
+	/* Check for a connected account */
+	connected_accounts = empathy_account_manager_get_connected_accounts (manager);
+
+	gtk_widget_set_sensitive (priv->message_item, connected_accounts > 0);
+}
+
 static void
 status_icon_finalize (GObject *object)
 {
@@ -350,11 +505,19 @@ status_icon_finalize (GObject *object)
 		g_source_remove (priv->blink_timeout);
 	}
 
-	empathy_disconnect_account_status_changed (priv->token);
+	g_signal_handlers_disconnect_by_func (priv->account_manager,
+					      status_icon_connection_changed_cb,
+					      object);
+
+	if (priv->notification) {
+		notify_notification_close (priv->notification, NULL);
+		g_object_unref (priv->notification);
+		priv->notification = NULL;
+	}
 
 	g_object_unref (priv->icon);
 	g_object_unref (priv->idle);
-	g_object_unref (priv->mc);
+	g_object_unref (priv->account_manager);
 	g_object_unref (priv->event_manager);
 }
 
@@ -368,33 +531,6 @@ empathy_status_icon_class_init (EmpathyStatusIconClass *klass)
 	g_type_class_add_private (object_class, sizeof (EmpathyStatusIconPriv));
 }
 
-static void
-status_icon_status_changed_cb (MissionControl           *mc,
-			       TpConnectionStatus        status,
-			       McPresence                presence,
-			       TpConnectionStatusReason  reason,
-			       const gchar              *unique_name,
-			       EmpathyStatusIcon        *icon)
-{
-	EmpathyStatusIconPriv *priv = GET_PRIV (icon);
-	GList                 *accounts, *l;
-	guint                  connection_status = 1;
-
-	/* Check for a connected account */
-	accounts = mc_accounts_list_by_enabled (TRUE);
-	for (l = accounts; l; l = l->next) {
-		connection_status = mission_control_get_connection_status (priv->mc,
-									   l->data,
-									   NULL);
-		if (connection_status == 0) {
-			break;
-		}
-	}
-	mc_accounts_list_free (accounts);
-
-	gtk_widget_set_sensitive (priv->message_item, connection_status == 0);
-}
-
 static void
 empathy_status_icon_init (EmpathyStatusIcon *icon)
 {
@@ -403,12 +539,13 @@ empathy_status_icon_init (EmpathyStatusIcon *icon)
 
 	icon->priv = priv;
 	priv->icon = gtk_status_icon_new ();
-	priv->mc = empathy_mission_control_new ();
-	priv->idle = empathy_idle_new ();
-	priv->event_manager = empathy_event_manager_new ();
-	priv->token = empathy_connect_to_account_status_changed (priv->mc,
-			G_CALLBACK (status_icon_status_changed_cb),
-			icon, NULL);
+	priv->account_manager = empathy_account_manager_dup_singleton ();
+	priv->idle = empathy_idle_dup_singleton ();
+	priv->event_manager = empathy_event_manager_dup_singleton ();
+
+	g_signal_connect (priv->account_manager,
+			  "account-connection-changed",
+			  G_CALLBACK (status_icon_connection_changed_cb), icon);
 
 	/* make icon listen and respond to MAIN_WINDOW_HIDDEN changes */
 	empathy_conf_notify_add (empathy_conf_get (),
@@ -428,6 +565,9 @@ empathy_status_icon_init (EmpathyStatusIcon *icon)
 	g_signal_connect (priv->event_manager, "event-removed",
 			  G_CALLBACK (status_icon_event_removed_cb),
 			  icon);
+	g_signal_connect (priv->event_manager, "event-updated",
+			  G_CALLBACK (status_icon_event_updated_cb),
+			  icon);
 	g_signal_connect (priv->icon, "activate",
 			  G_CALLBACK (status_icon_activate_cb),
 			  icon);
